from .cross_domain_classifier import SimpleCrossDomainClassifier
from .maml import MAML, CrossDomainMAML
from .protonet import ProtoNet, SimpleCrossDomainProtoNet
from .simple import SimpleClassifier
