from .cross_domain import CrossDomainModel, SimpleCrossDomainClassifier
from .episodic import EpisodicModel
from .maml import MAML, CrossDomainMAML
from .protonet import ProtoNet, SimpleCrossDomainProtoNet
from .simple import SimpleClassifier
